/*
 ============================================================================
 Name        : FESR-PiPhot-Regge.c
 Author      : Vincent Mathieu
 Version     :
 Copyright   : Your copyright notice
 Description : Hello World in C, Ansi-style
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <complex.h>    			/* Standard Library of Complex Numbers */


// UNITS = GeV
#define HBC     0.19732				// h-bar c
#define MP      0.93827203			// proton mass
#define MPI     0.13803899			// do not confuse MPI (pion mass) and M_PI (math pi)
#if !defined( M_PI )
	#define M_PI 3.141592654		// otherwise it doesn't compile on FuturGrid
#endif


// High Energy Routines
void FESR_Regge(
		double t, double massFin, int k[2],
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4], double fesr[13]);
void Ai_Regge(
		double t, double s,
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4], double complex Ai[13]);
void DSig_Regge(double t, double s,
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4],
		double Sig[4]);
void Pol_Regge(double t, double s,
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4],
		double Pol[4][3]);

double xsin(double x);

int main(void) {
	clock_t begin = clock();

	/* READ SIMULATION PARAMETERS. The format is
	 * Elab tmin tmax dt cutoff k1 k2 wmin wmax dw t0
	 * Elab				Lab energy for printing the observables for high energy
	 * tmin tmax dt		interval in t for the FESR
	 * wmin wmax dw		interval in w for printing Ai
	 * t0				print Ai at t = t0
	 * cutoff			W max for the FESR
	 * k1 k2			moments of the FESR
	 */

	double parSimu[20] = {0};
	FILE *fileParSimu = fopen("Simulation-param.txt","r");
	for (int i = 0; i<20; i++){ fscanf(fileParSimu, "%lf", &parSimu[i]); }
	fclose(fileParSimu);

	double Elab = parSimu[0];
	double s, t, tdeb = parSimu[1], tfin = parSimu[2], dt = parSimu[3], cut = parSimu[4];
	//double wdeb = parSimu[7], wfin = parSimu[8], dw = parSimu[9], t0  = parSimu[10]; 	// not needed in this code


	s = MP*MP + 2*MP*Elab;

	// k[0] needs to be odd ; k[1] needs to be even
	int k1 = parSimu[5], k2 = parSimu[6], k[2];
	k[0] = floor((k1-1)/2.0)*2+1;
	k[1] = floor(k2/2.0)*2;

	/* READ REGGE PARAMETERS
	 * 		parAlpha[i][0:1] = { alpha(0) alpha' }
	 * 		parRes[i][0:6]   = { j kappa delta beta b gamma1 gamma2 }
	 * There are 12 amplitudes with (max) 2 Regge exchanges --> 24 parameters
	 */
	double parAlpha[11][2] 	= {{0}};
	int    parRes1[25][3] 	= {{0}};
	double parRes2[25][4] 	= {{0}};
	FILE *fileAlpha = fopen("Trajectories.txt","r");
	FILE *fileRes   = fopen("Residues.txt","r");
	for (int i = 1; i<11; i++){ fscanf(fileAlpha, "%lf %lf", &parAlpha[i][0], &parAlpha[i][1]); }
	for (int i = 1; i<25; i++){
		fscanf(fileRes, "%i %i %i %lf %lf %lf %lf",
				&parRes1[i][0], &parRes1[i][1], &parRes1[i][2],
				&parRes2[i][0], &parRes2[i][1], &parRes2[i][2], &parRes2[i][3]);
	}
	fclose(fileAlpha);
	fclose(fileRes);


	// WRITE FESR FROM REGGE IN OUTPUT FILES
	FILE *fileFESR_High_All = fopen("FESR-High-All.txt","w");
	FILE *fileFESR_High_Pi0P = fopen("FESR-High-Pi0P.txt","w");
	double fesr_Reg[13] = {0};
	for(t = tdeb ; t > tfin  ; t += dt){
		FESR_Regge(t, cut, k, parAlpha, parRes1, parRes2, fesr_Reg);
		fprintf(fileFESR_High_All,
				"%8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n", t,
							fesr_Reg[1], fesr_Reg[2], fesr_Reg[3], fesr_Reg[4],
							fesr_Reg[5], fesr_Reg[6], fesr_Reg[7], fesr_Reg[8],
							fesr_Reg[9], fesr_Reg[10], fesr_Reg[11], fesr_Reg[12]);
		fprintf(fileFESR_High_Pi0P,
				"%8.4f %8.4f %8.4f %8.4f %8.4f \n", t,
							fesr_Reg[1] + fesr_Reg[2], fesr_Reg[4]  + fesr_Reg[5],
							fesr_Reg[7] + fesr_Reg[8], fesr_Reg[10] + fesr_Reg[11]);
	}
	fclose(fileFESR_High_All);	fclose(fileFESR_High_Pi0P);



	// WRITE DIFF. CROSS SECTION IN OUTPUT FILE
	// High Energy Limit Formulas
	FILE *fileDSig_High = fopen("DSig-High-HEL.txt","w");
	double Sig_Reg[4] = {0};
	for(t = tdeb ; t > tfin  ; t += dt){
		if(s < (MP+MPI)*(MP+MPI) ){ break; }
		DSig_Regge(t, s, parAlpha, parRes1, parRes2, Sig_Reg);
		fprintf(fileDSig_High,
			"%8.4f %8.4f %8.4f %8.4f %8.4f \n", t,
			Sig_Reg[0], Sig_Reg[1], Sig_Reg[2], Sig_Reg[3]);
	}
	fclose(fileDSig_High);


	// WRITE POL. OBSERVABLES IN OUTPUT FILE
	// High Energy Limit Formulas
	FILE *filePol_High_Pi0P = fopen("Pol-High-HEL-Pi0P.txt","w");
	FILE *filePol_High_Pi0N = fopen("Pol-High-HEL-Pi0N.txt","w");
	FILE *filePol_High_PiPN = fopen("Pol-High-HEL-PiPN.txt","w");
	FILE *filePol_High_PiMP = fopen("Pol-High-HEL-PiMP.txt","w");
	double Pol_Reg[4][3] = {{0}};
	for(t = tdeb ; t > tfin  ; t += dt){
		if(s < (MP+MPI)*(MP+MPI) ){ break; }
		Pol_Regge(t, s, parAlpha, parRes1, parRes2, Pol_Reg);

		fprintf(filePol_High_Pi0P, "%8.4f %8.4f %8.4f %8.4f \n", t,
			Pol_Reg[0][0], Pol_Reg[0][1], Pol_Reg[0][2]);

		fprintf(filePol_High_Pi0N, "%8.4f %8.4f %8.4f %8.4f \n", t,
			Pol_Reg[1][0], Pol_Reg[1][1], Pol_Reg[1][2]);

		fprintf(filePol_High_PiPN, "%8.4f %8.4f %8.4f %8.4f \n", t,
			Pol_Reg[2][0], Pol_Reg[2][1], Pol_Reg[2][2]);

		fprintf(filePol_High_PiMP, "%8.4f %8.4f %8.4f %8.4f \n", t,
			Pol_Reg[3][0], Pol_Reg[3][1], Pol_Reg[3][2]);
	}
	fclose(filePol_High_Pi0P);fclose(filePol_High_Pi0N);
	fclose(filePol_High_PiPN);fclose(filePol_High_PiMP);




	clock_t end = clock();
	double time_spent = (double)(end - begin) / CLOCKS_PER_SEC;
	printf("\n");
	printf("time spent: %f s", time_spent);


	return EXIT_SUCCESS;
}


// ----------------------- HIGH ENERGY  -------------------------------

// ********************************************************************
void Pol_Regge(double t, double s,
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4],
		double Pol[4][3]){
	/* Return the diff. cross section for the Regge contribution
	 * Pol[0][]: gamma p --> pi0 p
	 * Pol[1][]: gamma n --> pi0 n
	 * Pol[2][]: gamma p --> pi+ n
	 * Pol[3][]: gamma n --> pi- p
	 *
	 * Pol[][0]: Beam   asymmetry
	 * Pol[][1]: Target asymmetry
	 * Pol[][2]: Recoil asymmetry
	 */

	double no = 1, ba = 1, ta = 1, ra =1 ;
	// build the invariant amplitudes
	double complex Ai[13] = {0};
	double complex A1 = 0, A2 = 0, A3 = 0, A4 = 0;
	Ai_Regge(t, s, parAlpha, parRes1, parRes2, Ai);

	//gamma p --> pi0 p: A^0 + A^+
	A1 = Ai[1] + Ai[2];	A2 = Ai[4] + Ai[5];
	A3 = Ai[7] + Ai[8];	A4 = Ai[10] + Ai[11];

	no = A1*conj(A1) + A2*conj(A2) - t*( A4*conj(A4) + A3*conj(A3) );
	ba = A1*conj(A1) - A2*conj(A2) - t*( A4*conj(A4) - A3*conj(A3) );
	ta = 2*sqrt(-t)*cimag( A1*conj(A4) + A3*conj(A2) );
	ra = 2*sqrt(-t)*cimag( A1*conj(A4) - A3*conj(A2) );
	Pol[0][0] = ba/no; Pol[0][1] = ta/no; Pol[0][2] = ra/no;


	//gamma n --> pi0 n: -A^0 + A^+
	A1 = -Ai[1] + Ai[2];	A2 = -Ai[4] + Ai[5];
	A3 = -Ai[7] + Ai[8];	A4 = -Ai[10] + Ai[11];

	no = A1*conj(A1) + A2*conj(A2) - t*( A4*conj(A4) + A3*conj(A3) );
	ba = A1*conj(A1) - A2*conj(A2) - t*( A4*conj(A4) - A3*conj(A3) );
	ta = 2*sqrt(-t)*cimag( A1*conj(A4) + A3*conj(A2) );
	ra = 2*sqrt(-t)*cimag( A1*conj(A4) - A3*conj(A2) );
	Pol[1][0] = ba/no; Pol[1][1] = ta/no; Pol[1][2] = ra/no;


	//gamma p --> pi+ n: sqrt(2)*(A^0 + A^-)
	A1 = sqrt(2)*(Ai[1] + Ai[3]);	A2 = sqrt(2)*(Ai[4] + Ai[6]);
	A3 = sqrt(2)*(Ai[7] + Ai[9]);	A4 = sqrt(2)*(Ai[10] + Ai[12]);

	no = A1*conj(A1) + A2*conj(A2) - t*( A4*conj(A4) + A3*conj(A3) );
	ba = A1*conj(A1) - A2*conj(A2) - t*( A4*conj(A4) - A3*conj(A3) );
	ta = 2*sqrt(-t)*cimag( A1*conj(A4) + A3*conj(A2) );
	ra = 2*sqrt(-t)*cimag( A1*conj(A4) - A3*conj(A2) );
	Pol[2][0] = ba/no; Pol[2][1] = ta/no; Pol[2][2] = ra/no;

	//gamma n --> pi- p: sqrt(2)*(A^0 - A^-)
	A1 = sqrt(2)*(Ai[1] - Ai[3]);	A2 = sqrt(2)*(Ai[4] - Ai[6]);
	A3 = sqrt(2)*(Ai[7] - Ai[9]);	A4 = sqrt(2)*(Ai[10] - Ai[12]);

	no = A1*conj(A1) + A2*conj(A2) - t*( A4*conj(A4) + A3*conj(A3) );
	ba = A1*conj(A1) - A2*conj(A2) - t*( A4*conj(A4) - A3*conj(A3) );
	ta = 2*sqrt(-t)*cimag( A1*conj(A4) + A3*conj(A2) );
	ra = 2*sqrt(-t)*cimag( A1*conj(A4) - A3*conj(A2) );
	Pol[3][0] = ba/no; Pol[3][1] = ta/no; Pol[3][2] = ra/no;

	return;
}

// ********************************************************************
void DSig_Regge(double t, double s,
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4],
		double Sig[4]){
	/* Return the diff. cross section for the Regge contribution
	 * Sig[0]: gamma p --> pi0 p
	 * Sig[1]: gamma n --> pi0 n
	 * Sig[2]: gamma p --> pi+ n
	 * Sig[3]: gamma n --> pi- p
	 */

	// build the invariant amplitudes
	double complex Ai[13] = {0};
	Ai_Regge(t, s, parAlpha, parRes1, parRes2, Ai);

	Sig[0] =    		(Ai[1]+Ai[2])*conj(Ai[1]+Ai[2]) +   (Ai[4]+Ai[5])*conj(Ai[4]+Ai[5]);
	Sig[0] = Sig[0] - t*(Ai[7]+Ai[8])*conj(Ai[7]+Ai[8]) -t* (Ai[10]+Ai[11])*conj(Ai[10]+Ai[11]);
	Sig[0] = Sig[0]/32/M_PI*HBC*HBC*10000;

	Sig[1] =    		(Ai[1]-Ai[2])*conj(Ai[1]-Ai[2]) +   (Ai[4]-Ai[5])*conj(Ai[4]-Ai[5]);
	Sig[1] = Sig[1] - t*(Ai[7]-Ai[8])*conj(Ai[7]-Ai[8]) -t* (Ai[10]-Ai[11])*conj(Ai[10]-Ai[11]);
	Sig[1] = Sig[1]/32/M_PI*HBC*HBC*10000;

	Sig[2] =    		(Ai[1]+Ai[3])*conj(Ai[1]+Ai[3]) +   (Ai[4]+Ai[6])*conj(Ai[4]+Ai[6]);
	Sig[2] = Sig[2] - t*(Ai[7]+Ai[9])*conj(Ai[7]+Ai[9]) -t* (Ai[10]+Ai[12])*conj(Ai[10]+Ai[12]);
	Sig[2] = Sig[2]/32/M_PI*HBC*HBC*10000*sqrt(2.0);

	Sig[3] =    		(Ai[1]-Ai[3])*conj(Ai[1]-Ai[3]) +   (Ai[4]-Ai[6])*conj(Ai[4]-Ai[6]);
	Sig[3] = Sig[3] - t*(Ai[7]-Ai[9])*conj(Ai[7]-Ai[9]) -t* (Ai[10]-Ai[12])*conj(Ai[10]-Ai[12]);
	Sig[3] = Sig[3]/32/M_PI*HBC*HBC*10000*sqrt(2.0);

	return;
}

// ********************************************************************
void Ai_Regge(double t, double s,
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4],
		double complex Ai[13]){
	/* Return the invariant amplitudes Ai for the Regge contribution
	 * each Ai is a sum of two Regge contribution
	 * The 12 Ai are:
	 * 		1-3:   A_1^{0,+,-}
	 * 		4-6:   A_2^{0,+,-}
	 * 		7-9:   A_3^{0,+,-}
	 * 		10-12: A_4^{0,+,-}
	 */

	// build the trajectories
	double alpha[11] = {0};
	for(int i=1; i<11 ; i++){ alpha[i] = parAlpha[i][0] + parAlpha[i][1]*t;		}

	// build the residues
	double beta[25]   = {0};
	for(int i=1; i<25 ; i++){
		beta[i] = pow(t,parRes1[i][2]) ;
		beta[i] = beta[i]*parRes2[i][0]*exp(t*fabs(parRes2[i][1]));
		beta[i] = beta[i]*(1-t*parRes2[i][2])*(1-t*parRes2[i][3]);
	}

	// build the signature factors
	double complex signfac[11][3]   = {{1}};
	for(int i=1; i<11 ; i++){
		signfac[i][2] = (-1 - cexp(-I*M_PI*alpha[i]))/sin(M_PI*alpha[i]) ; // only for the pion
		signfac[i][1] = (-1 - cexp(-I*M_PI*alpha[i]))*xsin(alpha[i]) ; // for even amplitudes
		signfac[i][0] = I + tan(M_PI/2.0*alpha[i]) ; // for odd amplitudes
		}

	// even amplitudes for i = 1,2,4,5,9,10,11
	// odd  amplitudes for i = 3,6,7,8,12
	int mom[13] = {0,0,0,1,0,0,2,1,1,0,0,0,1} ;

	// build the FESR - there are 2 Regge contributions
	double nu = (s - MP*MP)/(2*MP) + (t-MPI*MPI)/(4*MP);
	double complex res1, res2;
	for(int i=1; i<13 ; i++){
		res1 = beta[2*i-1]*pow(nu, alpha[parRes1[2*i-1][0]] - 1);
		res1 = res1*signfac[parRes1[2*i-1][0]][mom[i]];

		res2 = beta[2*i]*pow(nu, alpha[parRes1[2*i][0]] - 1);
		res2 = res2*signfac[parRes1[2*i][0]][mom[i]];

		Ai[i] = res1 + res2 ;
	}
	return;
}

// ********************************************************************
void FESR_Regge(double t, double massFin, int k[2],
		double parAlpha[11][2], int parRes1[25][3], double parRes2[25][4],
		double fesr[13]){
	/* Return the Regge side of the FESR
	 * each FESR is a sum of two Regge contribution
	 * The 12 FESR are:
	 * 		1-3:  A_1^{0,+,-}
	 * 		4-6:  A_2^{0,+,-}
	 * 		7-9:  A_3^{0,+,-}
	 * 		9-12: A_4^{0,+,-}
	 *
	 * 		massFin is the cutoff, the mass in GeV
	 * 		k are the moments
	 * 		k[0] is odd and k[1] is even
	 */

	// build the trajectories
	double alpha[11] = {0};
	for(int i=1; i<11 ; i++){ alpha[i] = parAlpha[i][0] + parAlpha[i][1]*t;		}

	// build the residues
	double beta[25]   = {0};
	for(int i=1; i<25 ; i++){
		beta[i] = pow(alpha[parRes1[i][0]], parRes1[i][1])*pow(t,parRes1[i][2]) ;
		beta[i] = beta[i]*parRes2[i][0]*exp(t*fabs(parRes2[i][1]));
		beta[i] = beta[i]*(1-t*parRes2[i][2])*(1-t*parRes2[i][3]);
	}


	// odd  moments for i = 1,2,4,5,9,10,11
	// even moments for i = 3,6,7,8,12
	int mom[13] = {0,0,0,1,0,0,1,1,1,0,0,0,1} ;

	// build the FESR - there are 2 Regge contributions
	double Lam = (massFin*massFin - MP*MP)/(2*MP) + (t-MPI*MPI)/(4*MP);
	for(int i=1; i<13 ; i++){
		fesr[i] = beta[2*i-1]*pow(Lam, alpha[parRes1[2*i-1][0]] - 1)/(alpha[parRes1[2*i-1][0]] + k[mom[i]]);
		fesr[i] = fesr[i] + beta[2*i]*pow(Lam, alpha[parRes1[2*i][0]] - 1)/(alpha[parRes1[2*i][0]] + k[mom[i]]);
	}

	return;
}



// ********************************************************************
double xsin(double x){
	// return a Taylor expansion of x/sin(pi*x) around x = 0
	double taylor = 0;
	double y = M_PI*x/2.0;
	taylor = 1 + pow(y, 2)*2/3 + pow(y, 4)*14/45 + pow(y, 6)*124/945;
	taylor = taylor + pow(y, 8)*254/4725 + pow(y, 10)*292/13365 + pow(y, 12)*5657908/638512875;
	taylor = taylor + pow(y, 14)*65528/18243225 + pow(y, 16)*33862354/23260111875 + pow(y, 18)*22998766228/38979295480125;
	taylor = taylor + pow(y, 20)*366185109428/1531329465290625 + pow(y, 22)*26598037736/274446060013125 ;
	//taylor = taylor + pow(y, 24)*0.000039278321542146152 + pow(y, 26)*0.00001591890483211879;
	//taylor = taylor + pow(y, 28)*6.4516891916210/1000000 + pow(y, 30)*2.6147711488555591/1000000;

	return taylor/M_PI;
}
